// Copyright (C) 2021 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#ifndef QTQUICKPARTICLESEXPORTS_P_H
#define QTQUICKPARTICLESEXPORTS_P_H

// This file is autogenerated. Changes will be overwritten.

//
//  W A R N I N G
//  -------------
//
// This file is not part of the Qt API.  It exists purely as an
// implementation detail.  This header file may change from version to
// version without notice, or even be removed.
//
// We mean it.
//

#include <QtQuickParticles/qtquickparticlesexports.h>

#define Q_QUICKPARTICLES_PRIVATE_EXPORT Q_QUICKPARTICLES_EXPORT

#endif // QTQUICKPARTICLESEXPORTS_P_H
